/*
 * Copyright (C) 2002 Josh A. Beam
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>

extern char *proxyhost;
extern unsigned short proxyport;

extern int read_byte(int);
extern int establish_connection(char[], unsigned short);

/* connect to hostname:port via a socks5 proxy; returns file descriptor */
int
socks5_connect_to(char *hostname, unsigned short port,
                  char *username, char *password)
{
	int i;
	int fd;
	struct hostent *host;
	char buf[515];
	unsigned char len;
	unsigned char atyp;

	if(!proxyhost)
		return -1;

	host = gethostbyname(proxyhost);
	if(!host) {
		fprintf(stderr, "Error: Unable to resolve hostname %s\n", proxyhost);
		return -1;
	}

	fd = establish_connection(host->h_addr, proxyport);
	if(fd == -1) {
		fprintf(stderr, "Error: Unable to connect to SOCKS5 server %s:%u\n", proxyhost, proxyport);
		return -1;
	}

	fprintf(stderr, "Connected to SOCKS5 server %s:%u\n", proxyhost, proxyport);

	buf[0] = 0x05;
	buf[1] = 0x01;
	if(username && password)
		buf[2] = 0x02;
	else
		buf[2] = 0x00;
	send(fd, buf, 3, 0);
	if(read_byte(fd) != 0x05 || read_byte(fd) != buf[2]) {
		close(fd);
		return -1;
	}

	if(username && password) {
		unsigned char tmplen;

		buf[0] = 0x01;
		len = (strlen(username) > 255) ? 255 : strlen(username);
		buf[1] = len;
		memcpy(buf + 2, username, len);

		tmplen = (strlen(password) > 255) ? 255 : strlen(password);
		buf[2 + len] = tmplen;
		memcpy(buf + 3 + len, password, tmplen);

		send(fd, buf, (3 + len + tmplen), 0);

		if(read_byte(fd) != 0x01 || read_byte(fd) != 0x00) {
			fprintf(stderr, "Error: SOCKS5 authentication failed\n");
			close(fd);
			return -1;
		}
	}

	buf[0] = 0x05;
	buf[1] = 0x01;
	buf[2] = 0x00;
	buf[3] = 0x03;
	len = (strlen(hostname) > 255) ? 255 : strlen(hostname);
	buf[4] = (len & 0xff);
	memcpy(buf + 5, hostname, len);
	buf[5 + len] = (port >> 8);
	buf[6 + len] = (port & 0xff);
	send(fd, buf, (7 + len), 0);
	if(read_byte(fd) != 0x05 || read_byte(fd) != 0x00) {
		close(fd);
		return -1;
	}

	read_byte(fd);
	atyp = read_byte(fd);
	if(atyp == 0x01) {
		for(i = 0; i < 4; i++)
			read_byte(fd);
	} else if(atyp == 0x03) {
		len = read_byte(fd);
		for(i = 0; i < len; i++)
			read_byte(fd);
	} else {
		close(fd);
		return -1;
	}
	for(i = 0; i < 2; i++)
		read_byte(fd);

	return fd;
}

/* read from a socks5 connection */
int
socks5_read(int fd, char *buf, int size)
{
	return read(fd, buf, size);
}

/* send buf over a socks5 connection */
int
socks5_send(int fd, char *buf, int size)
{
	return send(fd, buf, size, 0);
}
